@@ -1,50 +1,53 @@
 """
 User-related Brood operations
 """
-from datetime import datetime, timedelta
+import base64
+import json
 import logging
-from random import randint
 import re
-from typing import Any, cast, Callable, Dict, List, Optional, Set
+import sys
 import uuid
+from datetime import datetime, timedelta
+from random import randint
+from typing import Any, Callable, Dict, List, Optional, Set, Tuple, cast
 
 from passlib.context import CryptContext
 from sendgrid import SendGridAPIClient
 from sendgrid.helpers.mail import Mail
-from sqlalchemy.orm.base import PASSIVE_OFF
-import stripe  # type: ignore
-from sqlalchemy import func, or_, and_
-from sqlalchemy.orm.session import Session
+from sqlalchemy import and_, func, or_
+from sqlalchemy.exc import IntegrityError
+from sqlalchemy.orm import Query
 from sqlalchemy.orm.exc import MultipleResultsFound
+from sqlalchemy.orm.session import Session
+from web3login.auth import to_checksum_address, verify
+from web3login.exceptions import Web3VerificationError
 
-from . import data
-from . import exceptions
-from . import subscriptions
+from . import data, exceptions, subscriptions
 from .models import (
-    KVBrood,
-    User,
-    VerificationEmail,
-    Token,
+    Application,
     Group,
-    GroupUser,
     GroupInvite,
-    UserGroupLimit,
-    Role,
-    TokenType,
+    GroupUser,
+    KVBrood,
     ResetPassword,
+    Role,
     Subscription,
     SubscriptionPlan,
-    Application,
+    Token,
+    TokenType,
+    User,
+    UserGroupLimit,
+    VerificationEmail,
 )
 from .settings import (
-    BUGOUT_URL,
     BUGOUT_FROM_EMAIL,
-    SENDGRID_API_KEY,
+    BUGOUT_URL,
     DEFAULT_USER_GROUP_LIMIT,
-    group_invite_link_from_env,
+    MOONSTREAM_APPLICATION_ID,
+    SENDGRID_API_KEY,
     TEMPLATE_ID_BUGOUT_WELCOME_EMAIL,
     TEMPLATE_ID_MOONSTREAM_WELCOME_EMAIL,
-    MOONSTREAM_APPLICATION_ID,
+    group_invite_link_from_env,
 )
 
 logger = logging.getLogger(__name__)
@@ -121,6 +124,12 @@ class TokenNotFound(Exception):
     """
 
 
+class TokenNotActive(Exception):
+    """
+    Raised when token with the given parameters is not active.
+    """
+
+
 class GroupInvalidParameters(ValueError):
     """
     Raised when operations are applied to a group but invalid parameters are provided with which to
@@ -209,6 +218,7 @@ def user_as_json_dict(user: User) -> Dict[str, Any]:
         "username": user.username,
         "email": user.email,
         "normalized_email": user.normalized_email,
+        "web3_address": user.web3_address,
         "verified": user.verified,
         "created_at": str(user.created_at),
         "updated_at": str(user.updated_at),
@@ -406,45 +416,82 @@ def password_confirm(
 def create_user(
     session: Session,
     username: str,
-    email: str,
-    password: str,
+    email: Optional[str] = None,
+    password: Optional[str] = None,
+    signature: Optional[str] = None,
     autogenerated_user: bool = False,
     first_name: Optional[str] = None,
     last_name: Optional[str] = None,
     application_id: Optional[uuid.UUID] = None,
+    is_verify: bool = False,
 ) -> User:
     """
-    Creates a new user in the given database session and
-    returns the created user object.
+    Creates a new user in the given database session and returns
+    the created user object.
 
-    According with autogenerated_user var it create bugout user for Slack/Github installation or
-    normal user.
+    According with autogenerated_user var it create bugout user for
+    Slack/Github installation or normal user.
 
     Sessions are expected to be sqlalchemy Session objects:
     https://docs.sqlalchemy.org/en/13/orm/session_api.html#sqlalchemy.orm.session.Session
     """
+    # Check username correctness
     assert username != ""
-    assert email != ""
-    assert password != ""
-
-    # Username and email should be stored as lowercase strings in the database.
     username = username.lower()
-    normalized_email = normalize_email(email)
-
     verify_username(username)
-    verify_password_strength(password)
 
-    password_context = get_password_context()
-    password_hash = password_context.hash(password)
-    auth_type = "brood"
+    is_creation_allowed = False
+
+    # Regular user with password and email creation
+    normalized_email = None
+    password_hash = None
+    if email is not None and password is not None:
+        assert email != ""
+        assert password != ""
+
+        # Email should be stored as lowercase strings in the database
+        normalized_email = normalize_email(email)
+        verify_password_strength(password)
+        password_context = get_password_context()
+        password_hash = password_context.hash(password)
+
+        is_creation_allowed = True
+
+    # Web3 user with blockchain address creation
+    web3_address = None
+    if signature is not None:
+        payload_json = base64.decodebytes(signature.encode()).decode("utf-8")
+        payload = json.loads(payload_json)
+        verified = verify(
+            authorization_payload=payload,
+            application_to_check=str(application_id)
+            if application_id is not None
+            else "",
+        )
+        if not verified:
+            logger.info("Web3 registration verification error")
+            raise Web3VerificationError()
+        web3_address = payload.get("address")
+        if web3_address is None:
+            logger.error(
+                f"Web3 address in payload could not be None for user with username: {username}"
+            )
+            raise Exception()
+        web3_address = to_checksum_address(web3_address)
+
+        is_creation_allowed = True
+
+    if not is_creation_allowed:
+        logger.info("Signature or email with password should be specified")
+        raise UserInvalidParameters()
 
     user_object = User(
         username=username,
         email=email,
         normalized_email=normalized_email,
         password_hash=password_hash,
-        auth_type=auth_type,
-        verified=True if autogenerated_user else False,
+        web3_address=web3_address,
+        verified=True if autogenerated_user else is_verify,
         autogenerated=True if autogenerated_user else False,
         first_name=first_name,
         last_name=last_name,
@@ -457,33 +504,37 @@ def create_user(
         session.add(user_object)
         session.add(user_group_limit)
         session.commit()
-    except Exception as e:
-        logger.error(e)
-        raise UserAlreadyExists("This user already exists")
+    except IntegrityError:
+        logger.info(
+            f"User already exists with username: {username}, email: {email}, address: {web3_address if signature is not None else None}"
+        )
+        raise UserAlreadyExists()
+    except Exception:
+        logger.error(
+            f"Unable to add user with username: {username}, email: {email}, address: {web3_address if signature is not None else None}"
+        )
+        raise Exception()
 
     return user_object
 
 
-def get_user(
-    session: Session,
+def get_user_query(
+    query: Query,
     username: Optional[str] = None,
     email: Optional[str] = None,
     user_id: Optional[uuid.UUID] = None,
     application_id: Optional[uuid.UUID] = None,
-) -> User:
+    web3_address: Optional[uuid.UUID] = None,
+) -> Query:
     """
-    Get a user by username, email, or user_id. If more than one of those fields is provided, will
-    look for a user having ALL the given parameters.
-
-    Emails are first normalized and then the lookup is performed against the normalized_email column
-    of the users table.
+    Apply filters to user query depends on argument provided.
     """
-    if username is None and email is None and user_id is None:
+    if username is None and email is None and user_id is None and web3_address is None:
         raise UserInvalidParameters(
-            "In order to get user, at least one of username, email, or user_id must be specified"
+            "In order to get user, at least one of username, email, user_id or web3_address must be specified"
         )
 
-    query = session.query(User)
+    query = query.filter(User.application_id == application_id)
 
     if username is not None:
         username = cast(str, username)
@@ -498,16 +549,214 @@ def get_user(
     if user_id is not None:
         query = query.filter(User.id == user_id)
 
-    if application_id is not None:
-        query = query.filter(User.application_id == application_id)
+    if web3_address is not None:
+        query = query.filter(User.web3_address == web3_address)
 
-    user = query.first()
-    if user is None:
-        raise UserNotFound(
-            f"Did not find user with filters username={username}, normalized_email={normalized_email}"
+    return query
+
+
+def get_user(
+    session: Session,
+    username: Optional[str] = None,
+    email: Optional[str] = None,
+    user_id: Optional[uuid.UUID] = None,
+    application_id: Optional[uuid.UUID] = None,
+    web3_address: Optional[uuid.UUID] = None,
+) -> User:
+    """
+    Get a user by username, email, user_id or web3_address.
+    If more than one of those fields is provided, will look for a user having ALL the given parameters.
+    """
+    if username is None and email is None and user_id is None and web3_address is None:
+        raise UserInvalidParameters(
+            "In order to get user, at least one of username, email, user_id, or web3_address must be specified"
         )
 
-    return user
+    query = session.query(User)
+    query = get_user_query(
+        query=query,
+        username=username,
+        email=email,
+        user_id=user_id,
+        application_id=application_id,
+        web3_address=web3_address,
+    )
+
+    objects = query.all()
+    if len(objects) == 0:
+        raise UserNotFound(
+            f"Did not find user with filters username: {username}, email: {email}, "
+            f"id: {user_id}, application_id: {application_id} and web3_address: {web3_address}"
+        )
+    if len(objects) > 1:
+        raise Exception(
+            f"Too many results for user with filters username: {username}, email: {email}, "
+            f"id: {user_id} and application_id: {application_id} and web3_address: {web3_address}"
+        )
+
+    return objects[0]
+
+
+def get_user_with_groups(
+    session: Session,
+    username: Optional[str] = None,
+    email: Optional[str] = None,
+    user_id: Optional[uuid.UUID] = None,
+    application_id: Optional[uuid.UUID] = None,
+    web3_address: Optional[uuid.UUID] = None,
+) -> data.UserWithGroupsResponse:
+    """
+    Get a user by username, email, user_id or web3_address.
+    If more than one of those fields is provided, will look for a user having ALL the given parameters.
+    """
+    if username is None and email is None and user_id is None and web3_address is None:
+        raise UserInvalidParameters(
+            "In order to get user, at least one of username, email, user_id, or web3_address must be specified"
+        )
+
+    query = session.query(User, Group, GroupUser)
+    query = get_user_query(
+        query=query,
+        username=username,
+        email=email,
+        user_id=user_id,
+        application_id=application_id,
+        web3_address=web3_address,
+    )
+    query = (
+        query.join(GroupUser, GroupUser.user_id == User.id, isouter=True)
+        .join(Group, Group.id == GroupUser.group_id, isouter=True)
+        .group_by(User, Group, GroupUser)
+    )
+
+    objects = query.all()
+    if len(objects) == 0:
+        raise UserNotFound(
+            f"Did not find user with filters username: {username}, email: {email}, "
+            f"id: {user_id}, application_id: {application_id} and web3_address: {web3_address}"
+        )
+
+    user = objects[0][0]
+    groups = []
+    for object in objects:
+        group = object[1]
+        group_user = object[2]
+        # Skip if there are no groups
+        if group is None:
+            continue
+        if group_user.user_id != user.id:
+            logger.error(
+                f"Unexpected group id: {group.id} fetched for user with id: {user.id}"
+            )
+            raise Exception("Unexpected group for user")
+        groups.append(
+            data.GroupUserResponse(
+                group_id=group.id,
+                user_id=group_user.user_id,
+                user_type=group_user.user_type,
+                autogenerated=group.autogenerated,
+                group_name=group.name,
+                parent=group.parent,
+            )
+        )
+
+    return data.UserWithGroupsResponse(
+        user_id=user.id,
+        username=user.username,
+        first_name=user.first_name,
+        last_name=user.last_name,
+        email=user.email,
+        normalized_email=user.normalized_email,
+        verified=user.verified,
+        created_at=user.created_at,
+        updated_at=user.updated_at,
+        autogenerated=user.autogenerated,
+        application_id=user.application_id,
+        web3_address=user.web3_address,
+        groups=groups,
+    )
+
+
+def get_current_user_by_token(session: Session, token: uuid.UUID) -> Tuple[bool, User]:
+    """
+    Get user by its active token for authentication middleware.
+    """
+    query = (
+        session.query(Token.active, User)
+        .join(User, User.id == Token.user_id)
+        .filter(Token.id == token)
+    )
+    objects = query.all()
+    if len(objects) == 0:
+        raise TokenNotFound("Token not found")
+    if len(objects) > 1:
+        logger.error(f"Found several users with token id: {token}")
+        raise Exception("Found several users")
+
+    is_token_active = objects[0][0]
+    user = objects[0][1]
+
+    return is_token_active, user
+
+
+def get_current_user_with_groups_by_token(
+    session: Session, token: uuid.UUID
+) -> Tuple[bool, data.UserWithGroupsResponse]:
+    """
+    Extend user data with groups for authentication middleware.
+    """
+    query = (
+        session.query(Token.active, User, Group, GroupUser)
+        .join(User, User.id == Token.user_id)
+        .join(GroupUser, GroupUser.user_id == User.id, isouter=True)
+        .join(Group, Group.id == GroupUser.group_id, isouter=True)
+        .filter(Token.id == token)
+        .group_by(Token.active, User, Group, GroupUser)
+    )
+    objects = query.all()
+    if len(objects) == 0:
+        raise TokenNotFound("Token not found")
+
+    active_token = objects[0][0]
+    user = objects[0][1]
+    groups = []
+    for object in objects:
+        if object[1].id != user.id:
+            logger.error(f"Found several users with token id: {token}")
+            raise Exception("Found several users")
+        # Skip if there are no groups
+        if object[2] is None:
+            continue
+        if object[3].user_id != user.id:
+            logger.error(
+                f"Unexpected group id: {object[2].id} fetched for user with id: {user.id}"
+            )
+            raise Exception("Unexpected group in list")
+        groups.append(
+            data.GroupUserResponse(
+                group_id=object[2].id,
+                user_id=object[3].user_id,
+                user_type=object[3].user_type,
+                autogenerated=object[2].autogenerated,
+                group_name=object[2].name,
+                parent=object[2].parent,
+            )
+        )
+
+    return active_token, data.UserWithGroupsResponse(
+        user_id=user.id,
+        username=user.username,
+        first_name=user.first_name,
+        last_name=user.last_name,
+        email=user.email,
+        normalized_email=user.normalized_email,
+        verified=user.verified,
+        created_at=user.created_at,
+        updated_at=user.updated_at,
+        autogenerated=user.autogenerated,
+        application_id=user.application_id,
+        groups=groups,
+    )
 
 
 def update_user(
@@ -638,8 +887,8 @@ def send_welcome_email(user: User, application_id: Optional[uuid.UUID] = None) -
         logger.info(
             f"Welcome email successfully submitted to Sendgrid for user with id={user_id}"
         )
-    except Exception as e:
-        logger.error(f"Error sending welcome email {e}")
+    except Exception:
+        logger.error(f"Error sending welcome email to {user.email}")
         pass
 
 
@@ -858,11 +1107,19 @@ def get_token(session: Session, token: uuid.UUID) -> Token:
     """
     token_object = session.query(Token).filter(Token.id == token).first()
     if token_object is None:
-        raise TokenNotFound(f"Token not found with ID: {token}")
+        raise TokenNotFound(f"Token not found")
     token_object = cast(Token, token_object)
     return token_object
 
 
+def get_tokens(session: Session, user_id: uuid.UUID) -> List[Token]:
+    """
+    Retrieve the list of tokens for user.
+    """
+    objects = session.query(Token).filter(Token.user_id == user_id).all()
+    return objects
+
+
 def update_token(
     session: Session,
     token: uuid.UUID,
@@ -1151,6 +1408,7 @@ def create_group(
             current_user_autogenerated=True if user.autogenerated is True else False,
             user_type=Role.owner.value,
             username=user.username,
+            application_id=user.application_id,
         )
     except Exception as e:
         logger.error(e)
@@ -1160,7 +1418,7 @@ def create_group(
 
 
 def get_groups_for_user(
-    session: Session, user_id: uuid.UUID
+    session: Session, user_id: uuid.UUID, include_metrics: bool = True
 ) -> List[data.GroupUserResponse]:
     """
     Get list of groups current user belongs to.
@@ -1179,19 +1437,27 @@ def get_groups_for_user(
         .all()
     )
 
-    groups_response = [
-        data.GroupUserResponse(
-            group_id=group.id,
-            user_id=group.user_id,
-            user_type=group.user_type,
-            autogenerated=group.autogenerated,
-            group_name=group.name,
-            parent=group.parent,
-            num_users=get_num_users(session, group),
-            num_seats=get_num_seats(session, group),
+    groups_response = []
+
+    for group in groups:
+        num_users = None
+        num_seats = None
+        if include_metrics:
+            num_users = get_num_users(session, group)
+            num_seats = get_num_seats(session, group)
+
+        groups_response.append(
+            data.GroupUserResponse(
+                group_id=group.id,
+                user_id=group.user_id,
+                user_type=group.user_type,
+                autogenerated=group.autogenerated,
+                group_name=group.name,
+                parent=group.parent,
+                num_users=num_users,
+                num_seats=num_seats,
+            )
         )
-        for group in groups
-    ]
 
     return groups_response
 
@@ -1240,6 +1506,7 @@ def set_user_in_group(
     user_type: str,
     username: Optional[str] = None,
     email: Optional[str] = None,
+    application_id: Optional[uuid.UUID] = None,
 ) -> data.GroupUserResponse:
     """
     Associate user with group or change user type in group if you have permissions.
@@ -1256,9 +1523,9 @@ def set_user_in_group(
     if username is not None:
         username = cast(str, username)
         username = username.lower()
-        user = get_user(session, username=username)
+        user = get_user(session, username=username, application_id=application_id)
     else:
-        user = get_user(session, email=email)
+        user = get_user(session, email=email, application_id=application_id)
 
     group = get_group(session, group_id=group_id)
 
